import datetime
import os

BASEDIR = os.path.abspath(os.path.dirname(__file__))


class Config:
    SECRET_KEY = os.environ.get('SECRET_KEY', 'something you will never guess')
    SQLALCHEMY_COMMIT_ON_TEARDOWN = True
    SQLALCHEMY_TRACK_MODIFICATIONS = True
    TEMPLATES_AUTO_RELOAD = True


class DevelopmentConfig(Config):
    ENV = "development"
    DEBUG = True


class ProductionConfig(Config):
    ENV = "production"
    DEBUG = False
    PERMANENT_SESSION_LIFETIME = datetime.timedelta(days=15)  # default is 31 days


config = {
    'development': DevelopmentConfig,
    'testing': ProductionConfig,
    'production': ProductionConfig,
    'default': DevelopmentConfig
}



from flask import Flask
from flask_moment import Moment

from .configs import *  # import configurations
from .database import *  # import database session
from .logger import *  # import logger
from .models import *  # import all the models
from .scheduler import *  # import scheduler
from .scrapper import *  # import scrapper
from .views import *  # import all the view


def create_app():
    app = Flask(__name__, static_url_path="/")
    config_app(app)

    with app.app_context():
        Moment(app)
        init_db()
        register_blue(app)

        # init_logger()
        # init_scheduler(app)
    @app.teardown_appcontext
    def teardown_session(e):
        my_db.close()
        OT_spider.close()

    @app.errorhandler(404)
    def page_not_found(e):
        return render_template("error/404.html"), 404

    @app.errorhandler(403)
    def access_forbidden(e):
        return render_template("error/403.html"), 403

    @app.errorhandler(500)
    def internal_server_error(e):
        return render_template("error/500.html"), 500

    @app.before_request
    def filter_request():
        if request.method not in ALLOWED_METHODS:
            return "Bad method", 403
        ua = str(request.user_agent)
        if "Mozilla" not in ua or "Gecko" not in ua:
            return "No Scrappers!", 403

    @app.after_request
    def set_res_headers(response):
        response.headers["Server"] = "OurTieba"
        response.headers["X-Content-Type-Options"] = "nosniff"
        response.headers["X-Frame-Options"] = "sameorigin"
        return response

    @app.template_filter("index_format")
    def add_zeros(i, length):  # format index in photos.html
        """
        Pad zeros to i, and turn it into a string. The length is at least 2.
        :param i: int. Integer to pad.
        :param length: int. Base integer.
        :return: A padded string.

        For example,
        add_zeros(1, 2) -> "01";
        add_zeros(1, 12) -> "01";
        add_zeros(13, 101) -> "013".
        """
        return ("{:0>" + str(max(len(str(length)), 2)) + "d}").format(i)

    if app.config['ENV'] == 'development':
        @app.route("/test")
        def sql_test():
            A = Admin(password="root", aname="root", timestamp="2000-01-01 00:00:00")

            u1 = User(uname="U1", password="111", nickname="user1", timestamp="2020-12-25 10:00:00",
                      avatar="avatar/-2837748924044579792.png")
            u2 = User(uname="U2", password="222", nickname="user2", timestamp="2020-12-26 08:00:00")
            u3 = User(uname="U3", password="333", nickname="user3", timestamp="2020-12-27 19:00:00")
            u4 = User(uname="U4", password="444", nickname="user4", timestamp="2020-12-28 20:00:00")
            u5 = User(uname="U5", password="555", nickname="user5", timestamp="2020-12-29 13:00:00")
            u6 = User(uname="U6", password="666", nickname="user6", timestamp="2020-12-30 17:00:00",
                      banned=1, banDuration="2022-01-01 00:00:00")

            b1 = Board(name="B1", hot=100, viewCount=15, subscribeCount=1, timestamp="2020-12-30 09:00:00",
                       postCount=3, sticky_on_top=20, description="D1")
            b2 = Board(name="B2", hot=50, viewCount=5, timestamp="2020-12-31 09:00:00", postCount=1,
                       description="D2")
            b3 = Board(name="B3", hot=120, viewCount=25, subscribeCount=1, timestamp="2021-02-01 10:40:00",
                       postCount=5, description="D3")
            b4 = Board(name="B4", hot=0, timestamp="2020-08-30 18:00:00", postCount=0, description="D4")
            b5 = Board(name="B5", hot=80, viewCount=12, timestamp="2021-01-30 09:30:00", postCount=2,
                       sticky_on_top=10, description="D5")
            b6 = Board(name="B6", hot=70, viewCount=10, timestamp="2020-10-29 10:05:00", postCount=2,
                       description="D6")

            p1 = Post(Uid=1, Bid=1, title="P1", content="<p>111</p>", viewCount=40, timestamp="2020-12-30 10:00:00",
                      commentCount=12, LCT="2021-03-21 09:00:00", dislikeCount=1, text="111", available_floor=14)
            p2 = Post(Uid=2, Bid=1, title="P2", content="<p>222</p>", timestamp="2021-01-30 15:00:00", text="222")
            p3 = Post(Uid=3, Bid=1, title="P3", content="<p>333</p>", timestamp="2021-02-03 20:40:00", text="333")
            p4 = Post(Uid=5, Bid=2, title="P4", content="<p>444</p>", timestamp="2021-03-20 10:50:00", text="444")
            p5 = Post(Uid=1, Bid=3, title="P5", content="<p>555</p>", timestamp="2021-02-02 10:40:00", text="555")
            p6 = Post(Uid=3, Bid=3, title="P6", content="<p>666</p>", timestamp="2021-02-02 16:00:00", text="666")
            p7 = Post(Uid=4, Bid=3, title="P7", content="<p>777</p>", viewCount=1, timestamp="2021-03-10 10:40:00",
                      likeCount=1, text="777")
            p8 = Post(Uid=4, Bid=3, title="P8", content="<p>888</p>", timestamp="2021-03-12 06:02:00", text="888")
            p9 = Post(Uid=1, Bid=3, title="P9", content="<p>999</p>", timestamp="2021-03-15 19:30:00", text="999")
            p10 = Post(Uid=2, Bid=5, title="P10", content="<p>1010</p>", timestamp="2021-02-04 15:00:00",
                       text="1010")
            p11 = Post(Uid=5, Bid=5, title="P11", content="<p>1111</p>", viewCount=3,
                       timestamp="2021-03-22 23:00:00",
                       commentCount=1, LCT="2021-03-23 13:03:00", text="1111", available_floor=3)
            p12 = Post(Uid=2, Bid=6, title="P12", content="<p>1212</p>", timestamp="2020-11-06 08:42:00",
                       text="1212")
            p13 = Post(Uid=2, Bid=6, title="P13", content="<p>1313</p>", timestamp="2021-03-21 12:01:00",
                       text="1313")
            p14 = Post(Uid=2, Bid=1, title="P14", content="<p>1414</p>", timestamp="2021-03-21 12:01:10",
                       text="1414")
            p15 = Post(Uid=2, Bid=1, title="P15", content="<p>1515</p>", timestamp="2021-03-21 12:01:00",
                       text="1515")
            p16 = Post(Uid=2, Bid=1, title="P16", content="<p>1616</p>", timestamp="2021-03-21 12:01:00",
                       text="1616")
            p17 = Post(Uid=2, Bid=1, title="P17", content="<p>1717</p>", timestamp="2021-03-21 12:01:00",
                       text="1717")
            p18 = Post(Uid=2, Bid=1, title="P18", content="<p>1818</p>", timestamp="2021-03-21 12:01:00",
                       text="1818")
            p19 = Post(Uid=2, Bid=1, title="P19", content="<p>1919</p>", timestamp="2021-03-21 12:01:00",
                       text="1919")
            p20 = Post(Uid=2, Bid=1, title="Game Visual", content='<p>Found on official twitter account:<br/>'
                                                                  '<img class="OT_image" src="/cdn/photo/-3701751787780283978'
                                                                  '.jpeg" title="" alt=""/></p>',
                       medias=["photo/-3701751787780283978.jpeg"],
                       timestamp="2021-05-05 08:11:11", text='Found on official twitter account:')
            p21 = Post(Uid=1, Bid=1, title="Edel Lilie Last Bullet Remix", content='<p>Song in the video below. '
                                                                                   'Enjoy~<br/><iframe class="OT_video '
                                                                                   'OT_iframe" width="420" height="280" '
                                                                                   'src="/play?src=/cdn/video'
                                                                                   '/-7344473432041877746.webm"></iframe><br'
                                                                                   '/>Here&#39;s the cover:<br/><img '
                                                                                   'class="OT_image" '
                                                                                   'src="/cdn/photo/-1976482922365202782.jpeg'
                                                                                   '" title="" alt=""/></p>',
                       text="Song in the video below. Enjoy~Here's the cover:", timestamp="2021-05-06 13:06:35",
                       medias=["video/-7344473432041877746.webm", "photo/-1976482922365202782.jpeg"])

            c1 = Comment(Uid=6, Pid=1, content="<p>wtf</p>", timestamp="2021-01-01 02:00:00", dislikeCount=1,
                         text="wtf", floor=2)
            c2 = Comment(Uid=1, Pid=1, content="<p>c111</p>", timestamp="2021-01-01 09:00:00", text="c111", floor=3)
            c3 = Comment(Uid=2, Pid=1, content="<p>c222</p>", timestamp="2021-01-02 09:00:00", text="c222", floor=4)
            c4 = Comment(Uid=3, Pid=1, content="<p>c333</p>", timestamp="2021-01-05 09:00:00", text="c333", floor=5)
            c5 = Comment(Uid=4, Pid=1, content="<p>c444</p>", timestamp="2021-01-21 09:00:00", text="c444", floor=6)
            c6 = Comment(Uid=5, Pid=1, content="<p>c555</p>", timestamp="2021-02-13 09:00:00", text="c555", floor=7)
            c7 = Comment(Uid=2, Pid=1, content="<p>c666</p>", timestamp="2021-03-01 09:00:00", likeCount=1,
                         text="c666", floor=8)
            c8 = Comment(Uid=3, Pid=1, content="<p>c777</p>", timestamp="2021-03-05 09:00:00", text="c777", floor=9)
            c9 = Comment(Uid=4, Pid=1, content="<p>c888</p>", timestamp="2021-03-08 09:00:00", text="c888",
                         floor=10)
            c10 = Comment(Uid=5, Pid=1, content="<p>c999</p>", timestamp="2021-03-11 09:00:00", text="c999",
                          floor=11)
            c11 = Comment(Uid=1, Pid=1, content="<p>c1010</p>", timestamp="2021-03-20 09:00:00", text="c1010",
                          floor=12)
            c12 = Comment(Uid=2, Pid=1, content="<p>c1111</p>", timestamp="2021-03-21 09:00:00", text="c1111",
                          floor=13)
            c13 = Comment(Uid=5, Pid=11, content="<p>c1212</p>", timestamp="2021-03-23 13:03:00", text="c1212",
                          floor=2)

            r1 = Report(Uid=1, target="comment", targetId=1, reason="yin zhan", timestamp="2021-01-01 09:05:00")
            r2 = Report(Uid=2, target="comment", targetId=1, reason="yin zhan!", timestamp="2021-01-02 09:05:00")
            r3 = Report(Uid=5, target="post", targetId=5, reason="dunno", timestamp="2021-02-03 10:40:00")

            cs1 = CommentStatus(Uid=1, Cid=1, liked=0, disliked=1, lastModified="2021-01-01 09:04:00")
            cs2 = CommentStatus(Uid=2, Cid=7, liked=1, disliked=0, lastModified="2021-03-02 11:00:00")

            ps1 = PostStatus(Uid=6, Pid=1, liked=0, disliked=1, lastModified="2021-01-01 01:59:00")
            ps2 = PostStatus(Uid=5, Pid=7, liked=1, disliked=0, lastModified="2021-03-10 12:00:00")

            s1 = Subscription(Uid=1, Bid=1, subscribed=1, lastModified="2020-12-30 09:00:00")
            s2 = Subscription(Uid=4, Bid=3, subscribed=1, lastModified="2021-02-03 14:50:00")

            h1 = History(Uid=1, Pid=1, lastVisitTime="2021-04-01 16:00:59")
            h2 = History(Uid=1, Pid=21, lastVisitTime="2021-03-24 18:00:59")
            h3 = History(Uid=2, Pid=3, lastVisitTime="2021-02-04 08:45:32")

            my_db.add(A)
            for u in [u1, u2, u3, u4, u5, u6]:
                my_db.add(u)

            for b in [b1, b2, b3, b4, b5, b6]:
                my_db.add(b)

            for p in [p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15, p16, p17, p18, p19, p20,
                      p21]:
                my_db.add(p)

            for c in [c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, c11, c12, c13]:
                my_db.add(c)

            for r in [r1, r2, r3]:
                my_db.add(r)

            for (uid, rid) in zip([1, 2, 5], [1, 2, 3]):
                u = my_db.query(User, User.Uid == uid, first=True)
                r = my_db.query(Report, Report.Rid == rid, first=True)
                u.reports.append(r)

            for s in [cs1, cs2, ps1, ps2]:
                my_db.add(s)

            for sub in [s1, s2]:
                my_db.add(sub)

            for history in [h1, h2, h3]:
                my_db.add(history)

            return "success!", 200
    return app



import logging

from .database import *  # must import this line here, otherwise cannot get sqlalchemy's logger


class Logger:
    def __init__(self):
        self.engine = my_db.get_engine().name
        logging.basicConfig(filename="ourtieba/logs/test.log",
                            level=logging.DEBUG,
                            format='%(asctime)s [%(thread)d:%(threadName)s] [%(filename)s:%(module)s:%(funcName)s] '
                                   '[%(levelname)s]: %(message)s',
                            filemode='a',
                            datefmt='%Y-%m-%d %A %H:%M:%S'
                            )
        self.logger = logging.getLogger('sqlalchemy.engine')
        self.logger.setLevel(logging.INFO)
        logging.getLogger('apscheduler.executors.default').setLevel(logging.WARNING)


def init_logger():
    logger = Logger()
    return logger



import datetime
import hashlib
import time

from sqlalchemy import Column, Integer, String, DateTime, DECIMAL
from sqlalchemy.orm import relationship

from ..configs.macros import AVATAR_PATH
from ._tables import user_report_table
from ..database import my_db


class User(my_db.Base):
    __tablename__ = 'user'

    Uid = Column(Integer, primary_key=True)
    password = Column(String)
    uname = Column(String, unique=True)
    nickname = Column(String)
    avatar = Column(String, default=AVATAR_PATH+"default_avatar.jpg")  # retrieved by hashing (Uid + upload timestamp)
    timestamp = Column(DateTime, default=datetime.datetime.utcnow)  # time of account creation
    lastCheck = Column(Integer, default=time.time)  # time the user last check message
    # personal info
    gender = Column(String)
    phoneNumber = Column(String)
    email = Column(String)
    address = Column(String)
    dateOfBirth = Column(DateTime)
    # ban status
    banned = Column(Integer, default=0)  # 0=False, 1=True
    banDuration = Column(DateTime, default=datetime.datetime.utcnow)  # banned until

    posts = relationship("Post", back_populates="owner")
    comments = relationship("Comment", back_populates="comment_by")
    reports = relationship("Report", secondary=lambda: user_report_table, back_populates="report_by")
    status_comment = relationship("CommentStatus", back_populates="by_user")
    status_post = relationship("PostStatus", back_populates="by_user")
    subscriptions = relationship("Subscription", back_populates="by_user")
    view = relationship('History', back_populates="by_user")

    def __init__(self, password, uname, nickname=None, avatar=None, timestamp=None, lastCheck=None, gender=None,
                 phone_number=None, email=None, address=None, dateOfBirth=None, banned=None, banDuration=None):
        if isinstance(timestamp, str):
            timestamp = datetime.datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S")
        if isinstance(dateOfBirth, str):
            dateOfBirth = datetime.datetime.strptime(dateOfBirth, "%Y-%m-%d %H:%M:%S")
        if isinstance(banDuration, str):
            banDuration = datetime.datetime.strptime(banDuration, "%Y-%m-%d %H:%M:%S")
        self.password = hashlib.sha3_512(password.encode()).hexdigest()
        self.uname = uname
        self.nickname = nickname if nickname else uname
        self.avatar = avatar
        self.timestamp = timestamp
        self.lastCheck = lastCheck
        self.gender = gender
        self.phoneNumber = phone_number
        self.email = email
        self.address = address
        self.dateOfBirth = dateOfBirth
        self.banned = banned
        self.banDuration = banDuration

    def __repr__(self):
        return '<User %r>' % self.Uid



import datetime

from sqlalchemy import Column, Integer, String, DateTime, ForeignKey, PickleType
from sqlalchemy.orm import relationship

from ..database import my_db


class Comment(my_db.Base):
    __tablename__ = "comment"

    Cid = Column(Integer, primary_key=True)
    content = Column(String, nullable=False)
    floor = Column(Integer, nullable=False)  # floor number given by post's available_floor on creation by time order
    medias = Column(PickleType, default=[])  # parsed from "content" column (see api.add_comment)
    text = Column(String, default="")  # plain text in "content" column
    timestamp = Column(DateTime, default=datetime.datetime.utcnow)
    status = Column(Integer, default=0)  # 0=normal, 1=deleted(by user), 2=banned(by admin)
    likeCount = Column(Integer, default=0)
    dislikeCount = Column(Integer, default=0)
    Uid = Column(Integer, ForeignKey("user.Uid"))
    Pid = Column(Integer, ForeignKey("post.Pid", ondelete='CASCADE'))

    comment_by = relationship("User", back_populates="comments")
    comment_in = relationship("Post", back_populates="comments")
    status_by = relationship("CommentStatus", back_populates="on_comment", cascade='all, delete', passive_deletes=True)

    def __init__(self, Uid, Pid, content, floor, medias=None, text=None, timestamp=None, status=None, likeCount=None,
                 dislikeCount=None):
        if isinstance(timestamp, str):
            timestamp = datetime.datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S")
        self.Uid = Uid
        self.Pid = Pid
        self.content = content
        self.floor = floor
        self.medias = medias
        self.text = text
        self.timestamp = timestamp
        self.status = status
        self.likeCount = likeCount
        self.dislikeCount = dislikeCount

    def __repr__(self):
        return '<Comment %r>' % self.Cid



from flask import Blueprint, make_response, abort, send_file

from ..configs import *


cdn = Blueprint("cdn", __name__, url_prefix="/cdn")


@cdn.route("/<path:filename>")
def cdn_send_file(filename):
    """
    This function is used to show corresponding photo to the users based on 'filename'
    :param filename:
    :return: response
    """
    # check whether file exists
    if not os.path.isfile(CDN_ROOT_PATH + filename):
        abort(404)
    res = make_response(send_file("../" + CDN_ROOT_PATH + filename))
    res.headers["Cache-Control"] = "public, max-age=86400"
    res.headers.pop("Expires")
    return res


import hashlib

from flask import Blueprint, jsonify, render_template, request

from ..configs import *
from ..database import *
from ..models import *

admin_blue = Blueprint("admin", __name__, url_prefix="/admin")


@admin_blue.route("/")
def admin_hello():
    """
    This function is used to redirect to admin dashboard
    :return: redirect to another path of '/admin/dashboard'
    """
    return redirect("/admin/dashboard")


@admin_blue.route("/login")
def admin_login():
    """
    This function is used to show the admin login page
    :return: admin_login.html
    """
    return render_template("admin_login.html")


@admin_blue.route("/dashboard")
@admin_login_required
def admin_dashboard():
    """
    This function is used for logged admins to show their dashboard
    :return: admin_dashboard.html
    """
    page = request.args.get("page", "1")
    order = Report.timestamp.desc()

    match_reports = my_db.query(Report, Report.resolved == 0, order)
    num_reports = len(match_reports)
    num_page = (num_reports - 1) // PAGE_SIZE + 1
    page = 1 if not page.isnumeric() or int(page) <= 0 else int(page) if int(page) <= num_page else num_page
    reports = [{"Rid": r.Rid, "target": r.target, "target_ID": r.targetId, "reason": r.reason,
                "timestamp": r.timestamp, "Uid": r.Uid}
               for r in match_reports[(page - 1) * PAGE_SIZE:page * PAGE_SIZE]]
    data = {"num_match": num_reports, "num_page": num_page, "page": page, "reports": reports}
    return render_template("admin_dashboard.html", data=data)


@admin_blue.route("/auth/login", methods=["POST"])
def admin_auth_login():
    """
    This function is to evaluate the process of login of the admins
    :return: json information
    if the process is successful it will return status 1 otherwise it will return error message
    """
    aname = request.form.get("aname")
    password = request.form.get("password")
    if not aname or not password:
        return jsonify({"error": {"msg": "Invalid data."}, "status": 0})

    admin_result = my_db.query(Admin, Admin.aname == aname, first=True)
    if not admin_result:
        return jsonify({"error": {"msg": "Admin name Not Found"}, 'status': 0})
    encoded_password = hashlib.sha3_512(password.encode()).hexdigest()
    recorded_password = admin_result.password
    if encoded_password != recorded_password:
        return jsonify({"error": {"msg": "Incorrect password."}, 'status': 0})

    session["Aid"] = admin_result.Aid
    admin_info = {"nickname": admin_result.nickname, "avatar": admin_result.avatar}
    session["admin_info"] = admin_info
    # session.permanent = True
    return jsonify({'status': 1})


@admin_blue.route("/auth/logout")
@admin_login_required
def admin_logout():
    """
    This function is used for logged in admins to logout
    :return: redirect to another path of '/admin/login'
    """
    session.clear()
    return redirect("/admin/login")


@admin_blue.route("/board/delete", methods=["POST"])
@admin_login_required
def admin_board_delete():
    """
    This function is used for logged in admin to delete a board
    :return: json information
    if the deletion is successful it will return status 1 otherwise it will return error message
    """
    Bid = request.form.get("Bid")
    if not Bid or not Bid.isnumeric():
        return jsonify({"error": {"msg": "Invalid data."}, "status": 0})
    match_board = my_db.query(Board, and_(Board.Bid == Bid, Board.status == 0), first=True)
    if not match_board:
        return jsonify({"error": {"msg": "Board not found."}, "status": 0})

    my_db.update(Board, Board.Bid == Bid, values={"status": 2})
    # Then delete all corresponding posts and comments in that Bid
    for p in match_board.posts:
        for c in p.comments:
            my_db.update(Comment, Comment.Cid == c.Cid, values={"status": 2})
        my_db.update(Post, Post.Pid == p.Pid, values={"status": 2})
    return jsonify({'status': 1})


@admin_blue.route("/post/delete", methods=["POST"])
@admin_login_required
def admin_post_delete():
    """
    This function is used for logged in admin to delete a post
    :return: json information
    if the deletion is successful it will return status 1 otherwise it will return error message
    """
    Pid = request.form.get("Pid")
    if not Pid or not Pid.isnumeric():
        return jsonify({"error": {"msg": "Invalid data."}, "status": 0})
    match_post = my_db.query(Post, and_(Post.Pid == Pid, Post.status == 0), first=True)
    if not match_post:
        return jsonify({"error": {"msg": "Post not found."}, "status": 0})

    match_post.under.postCount -= 1
    my_db.update(Post, Post.Pid == Pid, values={"status": 2})
    # Then delete all corresponding data in other relating tables
    for c in match_post.comments:
        my_db.update(Comment, Comment.Pid == c.Pid, status={"status": 2})
    return jsonify({'status': 1})


@admin_blue.route("/comment/delete", methods=["POST"])
@admin_login_required
def admin_comment_delete():
    """
    This function is used for logged in admin to delete a comment
    :return: json information
    if the deletion is successful it will return status 1 otherwise it will return error message
    """
    Cid = request.form.get("Cid")
    if not Cid or not Cid.isnumeric():
        return jsonify({"error": {"msg": "Invalid data."}, "status": 0})
    match_comment = my_db.query(Comment, and_(Comment.Cid == Cid, Comment.status == 0), first=True)
    if not match_comment:
        return jsonify({"error": {"msg": "Comment not found."}, "status": 0})

    match_comment.comment_in.commentCount -= 1
    my_db.update(Comment, Comment.Cid == Cid, values={"status": 2})
    return jsonify({'status': 1})


@admin_blue.route("/board/restore", methods=["POST"])
@admin_login_required
def admin_board_restore():
    """
    This function is used for logged in admin to delete a board
    :return: json information
    if the deletion is successful it will return status 1 otherwise it will return error message
    """
    Bid = request.form.get("Bid")
    if not Bid or not Bid.isnumeric():
        return jsonify({"error": {"msg": "Invalid data."}, "status": 0})
    match_board = my_db.query(Board, and_(Board.Bid == Bid, Board.status.in_((1, 2))), first=True)
    if not match_board:
        return jsonify({"error": {"msg": "Board not found."}, "status": 0})

    my_db.update(Board, Board.Bid == Bid, values={"status": 0})
    # Then delete all corresponding posts and comments in that Bid
    for p in match_board.posts:
        for c in p.comments:
            my_db.update(Comment, Comment.Cid == c.Cid, values={"status": 0})
        my_db.update(Post, Post.Pid == p.Pid, values={"status": 0})
    return jsonify({'status': 1})


@admin_blue.route("/post/restore", methods=["POST"])
@admin_login_required
def admin_post_restore():
    """
    This function is used for logged in admin to delete a post
    :return: json information
    if the deletion is successful it will return status 1 otherwise it will return error message
    """
    Pid = request.form.get("Pid")
    if not Pid or not Pid.isnumeric():
        return jsonify({"error": {"msg": "Invalid data."}, "status": 0})
    match_post = my_db.query(Post, and_(Post.Pid == Pid, Post.status.in_((1, 2))), first=True)
    if not match_post:
        return jsonify({"error": {"msg": "Post not found."}, "status": 0})

    match_post.under.postCount += 1
    my_db.update(Post, Post.Pid == Pid, values={"status": 0})
    # Then delete all corresponding data in other relating tables
    for c in match_post.comments:
        my_db.update(Comment, Comment.Pid == c.Pid, status={"status": 0})
    return jsonify({'status': 1})


@admin_blue.route("/comment/restore", methods=["POST"])
@admin_login_required
def admin_comment_restore():
    """
    This function is used for logged in admin to delete a comment
    :return: json information
    if the deletion is successful it will return status 1 otherwise it will return error message
    """
    Cid = request.form.get("Cid")
    if not Cid or not Cid.isnumeric():
        return jsonify({"error": {"msg": "Invalid data."}, "status": 0})
    match_comment = my_db.query(Comment, and_(Comment.Cid == Cid, Comment.status.in_((1, 2))), first=True)
    if not match_comment:
        return jsonify({"error": {"msg": "Comment not found."}, "status": 0})

    match_comment.comment_in.commentCount += 1
    my_db.update(Comment, Comment.Cid == Cid, values={"status": 0})
    return jsonify({'status': 1})


@admin_blue.route("/user/ban", methods=["POST"])
@admin_login_required
def admin_user_ban():
    """
    This function is used for logged in admin to ban a user
    :return: json information
    if the ban is successful it will return status 1 otherwise it will return error message
    """
    Uid = request.form.get("Uid")
    days = request.form.get("days")
    if not Uid or not Uid.isnumeric() or not days or not days.isnumeric() or int(days) <= 0:
        return jsonify({"error": {"msg": "Invalid data."}, "status": 0})
    affected_row = my_db.update(User, User.Uid == Uid,
                                values={"banned": 1,
                                        "banDuration": datetime.datetime.utcnow() + datetime.timedelta(days=int(days))})
    if not affected_row:
        return jsonify({"error": {"msg": "User not found."}, "status": 0})
    return jsonify({'status': 1})


@admin_blue.route("/user/unban", methods=["POST"])
@admin_login_required
def admin_user_unban():
    """
    This function is user for logged in admin to unban a user
    :return: json information
    if the ban is successful it will return status 1 otherwise it will return error message
    """
    Uid = request.form.get("Uid")
    if not Uid or not Uid.isnumeric():
        return jsonify({"error": {"msg": "Invalid data."}, "status": 0})
    affected_row = my_db.update(User, User.Uid == Uid, values={"banned": 0})
    if not affected_row:
        return jsonify({"error": {"msg": "User not found."}, "status": 0})
    return jsonify({'status': 1})


@admin_blue.route("/report/resolve", methods=["POST"])
@admin_login_required
def admin_report_resolve():
    """
    This function is used for logged in admin to resolve a report
    :return: json information
    if the process of resolution is successful it will return status 1 otherwise it will return error message
    """
    Rid = request.form.get("Rid")
    if not Rid or not Rid.isnumeric():
        return jsonify({"error": {"msg": "Invalid data."}, "status": 0})
    affected_row = my_db.update(Report, Report.Rid == Rid, values={"resolved": 1})
    if not affected_row:
        return jsonify({"error": {"msg": "Report not found."}, "status": 0})
    return jsonify({'status': 1})


@admin_blue.route("/create")
@admin_login_required
def create_board_interface():
    Aid = session["Aid"]

    match_admin = my_db.query(Admin, Admin.Aid == Aid, first=True)
    admin_info = {"Aid": Aid, "nickname": match_admin.nickname, "avatar": match_admin.avatar}
    data = {"admin_info": admin_info}
    return render_template("create_board.html", data=data)


from ..models._tables import user_report_table
from ..models.admin import Admin
from ..models.board import Board
from ..models.comment import Comment
from ..models.comment_status import CommentStatus
from ..models.notification import Notification
from ..models.post import Post
from ..models.post_status import PostStatus
from ..models.report import Report
from ..models.subscription import Subscription
from ..models.user import User
from ..models.view_history import History



from flask import Blueprint, render_template, session, redirect

guest = Blueprint("guest", __name__)


@guest.route("/register")
def register_interface():
    """
    This function is used for the guests to register
    :return: register.html, which is the register page
    """
    if session.get("Uid"):
        return redirect("/")
    return render_template("register.html")


@guest.route("/login")
def login_interface():
    """
    This function is used for the guests to login
    :return: login.html, which is the login page
    """
    if session.get("Uid"):
        return redirect("/")
    return render_template("login.html")



import os.path

from flask import Blueprint, render_template, request, session, abort

from ..configs.macros import *
from ..database import *
from ..models import *
from ..scrapper import *

a_user = Blueprint("abstract_user", __name__)


@a_user.route("/")
def index():
    """
    This function is used to show the main page of our system with recommend boards and hot news
    :return: index.html, which is our main page
    """
    hot_articles = OT_spider.get_hot_news(num=RECOMMEND_NUM_NEWS, freq=NEWS_UPDATE_FREQUENCY)
    hot_news = [{"title": a["title"], "abstract": a["description"], "link": f"/redirect?link={a['url']}",
                 "img_src": a["urlToImage"]} for a in hot_articles]
    boards = my_db.query(Board, and_(Board.status == 0), order=Board.hot.desc())[:RECOMMEND_NUM_BOARD]
    recommend_boards = [{"Bid": b.Bid, "name": b.name, "hot": b.hot, "post_count": b.postCount} for b in boards]
    data = {"boards": recommend_boards, "news": hot_news}
    return render_template("index.html", data=data)


@a_user.route("/board/<int:Bid>")
def get_posts_in_board(Bid):
    """
    This function is used to show the corresponding board page according to Bid with several posts in it.
    :param Bid: the id of a Board
    :return: board.html, a board with corresponding Bid
    """
    b = my_db.query(Board, and_(Board.Bid == Bid, Board.status == 0), first=True)
    if not b:
        abort(404)

    b.viewCount += 1  # when page is accessed, increment view count
    subs = my_db.query(Subscription, and_(Subscription.Uid == (Uid := session.get("Uid")),
                                          Subscription.Bid == Bid), first=True)

    board_info = {"Bid": b.Bid, "name": b.name, "hot": b.hot, "post_count": b.postCount, "subs_count": b.subscribeCount,
                  "time": b.timestamp, "view_count": b.viewCount, "cover": b.cover, "description": b.description,
                  "subs_by_user": subs.subscribed if subs else 0}

    order = request.args.get("order", "latest_comment")
    page = request.args.get("page", "1")
    if order == "latest_comment":
        order = Post.latestCommentTime.desc()
    elif order == "newest":
        order = Post.timestamp.desc()
    elif order == "like_count":
        order = Post.likeCount.desc()
    else:
        order = Post.commentCount.desc()

    posts_match_result = my_db.query(Post, and_(Post.Bid == Bid, Post.status == 0), order)
    num_match = len(posts_match_result)
    num_page = (num_match - 1) // PAGE_SIZE + 1
    page = 1 if not page.isnumeric() or int(page) <= 0 else int(page) if int(page) <= num_page else num_page
    posts = []
    for p in posts_match_result[(page - 1) * PAGE_SIZE:page * PAGE_SIZE]:
        post_info = {"Pid": (Pid := p.Pid), "Uid": p.Uid, "title": p.title, "summary": p.text,
                     "publish_time": p.timestamp, "comment_count": p.commentCount, "like_count": p.likeCount,
                     "dislike_count": p.dislikeCount, "max_floor": p.available_floor,
                     "preview_type": None, "preview_src": None}
        if p.medias:
            preview_media = p.medias[0].split("/")
            post_info.update({"preview_type": preview_media[0], "preview_src": preview_media[1]})
        posts.append(post_info)

        if not session.get("Uid"):
            post_info.update({"liked_by_user": 0, "disliked_by_user": 0})
        else:
            Uid = session["Uid"]
            match_status = my_db.query(PostStatus, and_(PostStatus.Uid == Uid, PostStatus.Pid == Pid), first=True)
            if not match_status:
                post_info.update({"liked_by_user": 0, "disliked_by_user": 0})
            else:
                post_info.update({"liked_by_user": match_status.liked, "disliked_by_user": match_status.disliked})

    data = {"num_match": num_match, "num_page": num_page, "page": page, "posts": posts, "board_info": board_info}

    return render_template("board.html", data=data)


@a_user.route("/post/<int:Pid>")
def get_comments_in_post(Pid):
    """
    This function is used to show the corresponding post page according to Pid with several comments in it
    :param Pid: the id of a Post
    :return: post.html, a post with corresponding Pid
    """
    p = my_db.query(Post, and_(Post.Pid == Pid, Post.status == 0), first=True)
    if not p:
        abort(404)

    p.viewCount += 1  # when page is accessed, increment view count
    post_info = {"Pid": p.Pid, "Bid": p.Bid, "Uid": p.Uid, "title": p.title, "content": p.content,
                 "publish_time": p.timestamp, "comment_count": p.commentCount, "like_count": p.likeCount,
                 "dislike_count": p.dislikeCount, "owner": (o := p.owner).nickname, "avatar": o.avatar}
    board_info = {"cover": (b := p.under).cover, "bname": b.name}
    if not session.get("Uid"):
        post_info.update({"liked_by_user": 0, "disliked_by_user": 0})
    else:
        Uid = session["Uid"]
        match_status = my_db.query(PostStatus, and_(PostStatus.Uid == Uid, PostStatus.Pid == Pid), first=True)
        if not match_status:
            post_info.update({"liked_by_user": 0, "disliked_by_user": 0})
        else:
            post_info.update({"liked_by_user": match_status.liked, "disliked_by_user": match_status.disliked})

        match_history = my_db.query(History, and_(History.Uid == Uid, History.Pid == Pid), first=True)
        if not match_history:
            new_history = History(Uid, Pid)
            my_db.add(new_history)
        else:
            match_history.lastVisitTime = datetime.datetime.utcnow()

    order = request.args.get("order")
    page = request.args.get("page", "1")

    if order == "desc":
        order = Comment.timestamp.desc()
    elif order == "like_count":
        order = Comment.likeCount.desc()
    else:  # if order is None or invalid parameters
        order = Comment.timestamp  # default order is asc()

    comment_match_result = my_db.query(Comment, and_(Comment.Pid == Pid, Comment.status == 0), order)
    num_match = len(comment_match_result)
    num_page = (num_match - 1) // PAGE_SIZE + 1
    page = 1 if not page.isnumeric() or int(page) <= 0 else int(page) if int(page) <= num_page else num_page

    Comments = []
    for c in comment_match_result[(page - 1) * PAGE_SIZE:page * PAGE_SIZE]:
        base_info = {"Cid": c.Cid, "Uid": c.Uid, "content": c.content, "publish_time": c.timestamp,
                     "like_count": c.likeCount, "dislike_count": c.dislikeCount, "publish_user": c.comment_by.nickname,
                     "user_avatar": c.comment_by.avatar, "floor": c.floor}
        if not session.get("Uid"):
            base_info.update({"liked_by_user": 0, "disliked_by_user": 0})
        else:
            Uid = session["Uid"]
            match_status = my_db.query(CommentStatus, and_(CommentStatus.Uid == Uid, CommentStatus.Cid == c.Cid),
                                       first=True)
            if not match_status:
                base_info.update({"liked_by_user": 0, "disliked_by_user": 0})
            else:
                base_info.update({"liked_by_user": match_status.liked, "disliked_by_user": match_status.disliked})
        Comments.append(base_info)

    data = {"num_match": num_match, "num_page": num_page, "page": page, "comments": Comments, "post_info": post_info,
            "board_info": board_info}  # board here means the board that the post is under
    return render_template("post.html", data=data)


@a_user.route("/search_board")
def search_board():
    """
    This function is used to search for relating board based on the key words users give.
    :return: search_result.html, which shows all corresponding boards satisfying the demand
    """
    keyword = request.args.get("kw")
    if not keyword:
        return render_template("search_result.html", error="Please enter a keyword!")
    order = request.args.get("order", "popular")
    page = request.args.get("page", "1")
    order = Board.timestamp.desc() if order == "popular" else Board.hot.desc()
    match_result = my_db.query(Board, and_(Board.name.like("%" + keyword + "%"), Board.status == 0), order)
    num_match = len(match_result)
    num_page = (num_match - 1) // PAGE_SIZE + 1
    page = 1 if not page.isnumeric() or int(page) <= 0 else int(page) if int(page) <= num_page else num_page
    boards = [{"Bid": b.Bid, "name": b.name, "hot": b.hot, "post_count": b.postCount}
              for b in match_result[(page - 1) * PAGE_SIZE:page * PAGE_SIZE]]
    data = {"num_match": num_match, "num_page": num_page, "page": page, "boards": boards}
    return render_template("search_result.html", data=data)


@a_user.route("/profile/<int:Uid>")
def get_personal_profile(Uid):
    """
    This function is used to show the corresponding profile page according to Uid
    :param Uid:
    :return: profile.html, which contains many information of corresponding user
    """
    u = my_db.query(User, User.Uid == Uid, first=True)
    if not u:
        abort(404)

    post_count = my_db.count(Post, and_(Post.Uid == Uid, Post.status == 0))
    comment_count = my_db.count(Comment, and_(Comment.Uid == Uid, Comment.status == 0))
    subs_count = my_db.query_join(Subscription.Uid, Board, and_(Subscription.Uid == Uid, Subscription.subscribed == 1,
                                                                Board.status == 0), count=True)
    history_count = my_db.count(History, History.Uid == Uid)

    user_info = {
        "nickname": u.nickname, "avatar": u.avatar, "timestamp": u.timestamp, "gender": u.gender,
        "phoneNumber": u.phoneNumber, "email": u.email, "address": u.address, "dateOfBirth": u.dateOfBirth,
        "banned": u.banned, "banDuration": str(u.banDuration), "isCurrent": int(Uid == session.get("Uid", -1)),
        "post_count": post_count, "subs_count": subs_count, "history_count": history_count, "Uid": Uid,
        "comment_count": comment_count
    }
    return render_template("profile.html", data=user_info)


@a_user.route("/photos")
def photo_gallery():
    """
    This function is used to view the images in the board and post.
    :return: photos.html, which contains all photos related to this board or post, and
    the initial position of photo in the slides (not photo list)
    """
    # if only "Pid" param, show all images in post content, else if valid "src" param,
    # show all images in both post and comment content, else show nothing
    Pid = request.args.get("Pid")
    src = request.args.get("src")
    if not Pid:
        abort(404)
    match_post = my_db.query(Post, Post.Pid == Pid, first=True)
    if not match_post:
        abort(404)

    photos = []
    position = None
    # get photos in post
    medias = match_post.medias
    base_len = 0
    for i, m in enumerate(medias):
        if m.startswith(PHOTO_PATH):
            base_len += 1
            cur_src = "/" + CDN_ROOT_PATH + m
            photos.append(cur_src)
            if cur_src == src:
                position = base_len
    # also get photos in comment if "src" specified
    if src:
        comments = my_db.query(Comment, Comment.Pid == Pid)
        for c in comments:
            medias = c.medias
            for i, m in enumerate(medias):
                if m.startswith(PHOTO_PATH):
                    base_len += 1
                    cur_src = "/" + CDN_ROOT_PATH + m
                    photos.append(cur_src)
                    if cur_src == src:
                        position = base_len
        if position is None:  # which means invalid "src"
            abort(404)
    else:
        position = 1 if photos else 0
    data = {"photos": photos, "init_index": position}
    return render_template("photos.html", data=data)


@a_user.route("/redirect")
def redirect_page():
    """
    This function is used to redirect user when user clicks an external link. Valid links begins with "http://",
    "https://" or "ftp://".
    :return: redirect.html, which contains the link of the outside news
    """
    link: str = request.args.get("link")
    if not link or not (link.startswith("http://") or link.startswith("https://") or link.startswith("ftp://")):
        data = {"status": 0}
    else:
        data = {"link": link, "status": 1}
    return render_template("redirect.html", data=data)


# mainly for pasring url, can also do in "dplayer_embed.html" by pure javascript
@a_user.route("/play")
def render_dplayer():
    """
    This function is used to help the users to view videos
    :return: dplayer_embed.html, which can allow the users to view videos
    """
    src = request.args.get("src")
    if not src:
        abort(404)
    if not src.startswith("http"):  # inner src link
        src = os.path.join(CDN_ROOT_PATH, src)
    autoplay = request.args.get("autoplay")
    if not autoplay or not autoplay.isnumeric():
        autoplay = 0  # default is no autoplay
    loop = request.args.get("loop")
    if not loop or not loop.isnumeric():
        loop = 0  # default is no loop
    data = {"src": src, "autoplay": autoplay, "loop": loop}
    return render_template("dplayer_embed.html", data=data)



from sqlalchemy import Column, Integer, Table, ForeignKey

from ..database import my_db

user_report_table = Table("user_report", my_db.Base.metadata,
                          Column("Uid", Integer, ForeignKey("user.Uid")),
                          Column("Rid", Integer, ForeignKey("report.Rid"))
                          )



import datetime

from sqlalchemy import Column, Integer, String, DateTime, ForeignKey, PickleType
from sqlalchemy.orm import relationship

from ..database import my_db


class Post(my_db.Base):
    __tablename__ = "post"

    Pid = Column(Integer, primary_key=True)
    title = Column(String, nullable=False)
    content = Column(String, default="<p></p>")
    medias = Column(PickleType, default=[])  # parsed from "content" column (see api.add_post)
    text = Column(String, default="")  # plain text in "content" column
    timestamp = Column(DateTime, default=datetime.datetime.utcnow)
    # next available index for a floor in the post, increment on comment creation, DO NOT decrement on comment deletion
    available_floor = Column(Integer, default=2)  # available_floor - 1 is the number of comments ever posted in a post
    status = Column(Integer, default=0)  # 0=normal, 1=deleted(by user), 2=banned(by admin)
    sticky_on_top = Column(PickleType, default=[])  # the list of Cid of comment sticky on top
    commentCount = Column(Integer, default=0)
    likeCount = Column(Integer, default=0)
    dislikeCount = Column(Integer, default=0)
    viewCount = Column(Integer, default=0)
    latestCommentTime = Column(DateTime, default=datetime.datetime.utcnow)

    Uid = Column(Integer, ForeignKey("user.Uid"))
    Bid = Column(Integer, ForeignKey("board.Bid"))

    under = relationship("Board", back_populates="posts")
    owner = relationship("User", back_populates="posts")
    comments = relationship("Comment", back_populates="comment_in", cascade='all, delete', passive_deletes=True)
    status_by = relationship("PostStatus", back_populates="on_post", cascade='all, delete', passive_deletes=True)
    view = relationship("History", back_populates="related_post")

    def __init__(self, Uid, Bid, title, content=None, medias=None, text=None, timestamp=None, LCT=None, status=None,
                 viewCount=None, commentCount=None, likeCount=None, dislikeCount=None, available_floor=None):
        if isinstance(timestamp, str):
            timestamp = datetime.datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S")
        if isinstance(LCT, str):
            LCT = datetime.datetime.strptime(LCT, "%Y-%m-%d %H:%M:%S")
        self.Uid = Uid
        self.Bid = Bid
        self.title = title
        self.content = content
        self.medias = medias
        self.text = text
        self.timestamp = timestamp
        self.available_floor = available_floor
        self.status = status
        self.latestCommentTime = LCT
        self.viewCount = viewCount
        self.commentCount = commentCount
        self.likeCount = likeCount
        self.dislikeCount = dislikeCount

    def __repr__(self):
        return '<Post %r>' % self.Pid



import datetime

from sqlalchemy import Column, Integer, DateTime, ForeignKey
from sqlalchemy.orm import relationship

from ..database import my_db


class Subscription(my_db.Base):
    __tablename__ = "subscription"

    Uid = Column(Integer, ForeignKey("user.Uid"), primary_key=True)
    Bid = Column(Integer, ForeignKey("board.Bid"), primary_key=True)
    # Two implementations:
    # 1. Not use "subscribed" column. Simply delete from table when unsubscribe;
    # 2. Use "subscribed" column. Set subscribed = 0 when unsubscribe.
    subscribed = Column(Integer, default=0)  # 0 = False, 1 = True,
    lastModified = Column(DateTime, default=datetime.datetime.utcnow)  # timestamp of last action

    by_user = relationship("User", back_populates="subscriptions")
    of_board = relationship("Board", back_populates="subscribers")

    def __init__(self, Uid, Bid, subscribed=None, lastModified=None):
        if isinstance(lastModified, str):
            lastModified = datetime.datetime.strptime(lastModified, "%Y-%m-%d %H:%M:%S")
        self.Uid = Uid
        self.Bid = Bid
        self.subscribed = subscribed
        self.lastModified = lastModified

    def __repr__(self):
        return f"<Subscription ({self.Uid}, {self.Bid})>"


from contextlib import contextmanager
import datetime
import functools

from flask import session, redirect


# For now, we just assume that all the sessions are not tampered.
# Forgery may be possible, but it's csrf token's lob to find it out
# (which we haven't implemented yet).
def login_required(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if session.get("Uid"):
            return f(*args, **kwargs)
        else:
            return redirect("/login")

    return wrapper


def admin_login_required(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if session.get("Aid"):
            return f(*args, **kwargs)
        else:
            return redirect("/admin/login")

    return wrapper


@contextmanager
def auto_scope(_session):
    if not _session:
        raise Exception("Please connect to database first!")
    try:
        yield _session
        _session.commit()
    except Exception as e:
        _session.rollback()
        print(e)


def convert_time(ts: datetime.datetime):
    if ts.strftime("%Y") != datetime.datetime.utcnow().strftime("%Y"):
        return ts.strftime("%Y-%m-%d")
    if (day := ts.strftime("%m-%d")) != datetime.datetime.utcnow().strftime("%m-%d"):
        return day
    return "Today " + ts.strftime("%H:%M")


if __name__ == '__main__':
    new = datetime.datetime(2021, 5, 9, 15, 0, 0)
    print(convert_time(new))


import datetime

from sqlalchemy import Column, Integer, String, DateTime, PickleType
from sqlalchemy.orm import relationship

from ..database import my_db


class Board(my_db.Base):
    __tablename__ = "board"

    Bid = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    description = Column(String, nullable=False)
    hot = Column(Integer, default=0)
    cover = Column(String, default="cover/OurTieba.png")
    status = Column(Integer, default=0)  # 0=normal, 1=deleted(by user), 2=banned(by admin)
    sticky_on_top = Column(PickleType, default=[])  # the list of Pid of post sticky on top
    postCount = Column(Integer, default=0)
    viewCount = Column(Integer, default=0)
    subscribeCount = Column(Integer, default=0)
    timestamp = Column(DateTime, default=datetime.datetime.utcnow)

    posts = relationship("Post", back_populates="under")
    subscribers = relationship("Subscription", back_populates="of_board")

    def __init__(self, name, description, hot=None, cover=None, status=None, sticky_on_top=None, postCount=None,
                 viewCount=None, subscribeCount=None, timestamp=None, ):
        if isinstance(timestamp, str):
            timestamp = datetime.datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S")
        self.name = name
        self.description = description
        self.hot = hot
        self.cover = cover
        self.status = status
        self.sticky_on_top = sticky_on_top
        self.viewCount = viewCount
        self.subscribeCount = subscribeCount
        self.postCount = postCount
        self.timestamp = timestamp

    def __repr__(self):
        return "<Board %r>" % self.Bid


from sqlalchemy import create_engine, func, inspect, and_
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import scoped_session, sessionmaker

from .configs import auto_scope


class myDb:
    __instance = None

    Base = declarative_base()

    # ensure that only one instance is created
    def __new__(cls, *args, **kwargs):
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance

    def __init__(self):
        self._engine = None
        self._session = None

    def connect(self, *args, **kwargs):
        self._engine = create_engine('sqlite:///test.db', convert_unicode=True)
        # must use scoped session here
        self._session = scoped_session(sessionmaker(self._engine, *args, **kwargs))
        self.Base.query = self._session.query_property()

    def initialize(self, *args, **kwargs):
        self.connect(*args, **kwargs)
        # import all models here (only the class name)
        from .models import Admin, Board, Comment, CommentStatus, Post, PostStatus, Report, Subscription,\
            User, user_report_table
        self.Base.metadata.create_all(bind=self._engine)

    def query(self, target, condition=True, order=True, first=False):
        with auto_scope(self._session) as _db_session:
            if first:
                return _db_session.query(target).filter(condition).order_by(order).first()
            return _db_session.query(target).filter(condition).order_by(order).all() or []  # if None return []

    def query_join(self, target, join, condition=True, order=True, first=False, count=False):
        with auto_scope(self._session) as _db_session:
            if first:
                return _db_session.query(target).join(join).filter(condition).order_by(order).first()
            if count:
                return _db_session.query(func.count(target)).join(join).filter(condition).order_by(order).scalar()
            return _db_session.query(target).join(join).filter(condition).order_by(order).all() or []

    def add(self, new):
        with auto_scope(self._session) as _db_session:
            return _db_session.add(new)

    def merge(self, new):
        with auto_scope(self._session) as _db_session:
            return _db_session.merge(new)

    def delete(self, target, condition=False, synchronize_session="fetch"):
        with auto_scope(self._session) as _db_session:
            return _db_session.query(target).filter(condition).delete(synchronize_session)

    def avg(self, target, condition=True):
        with auto_scope(self._session) as _db_session:
            return _db_session.query(func.avg(target)).filter(condition).scalar()

    def count(self, target, condition=True):
        if issubclass(target, self.Base):
            target = inspect(target).primary_key[0]
        with auto_scope(self._session) as _db_session:
            return _db_session.query(func.count(target)).filter(condition).scalar()

    def update(self, target, condition=True, **kwargs):
        with auto_scope(self._session) as _db_session:
            values = kwargs.get("values")  # "values" is a dict
            if values:
                return _db_session.query(target).filter(condition).update(values)

    def close(self):
        self._session.remove()

    def get_engine(self):
        return self._engine

    def set_engine(self, _new_engine):
        self._engine = _new_engine


class dbFactory:
    @staticmethod
    def produce():
        return myDb()


my_db = dbFactory.produce()


def init_db():
    my_db.initialize()



import datetime

from sqlalchemy import Column, Integer, DateTime, ForeignKey
from sqlalchemy.orm import relationship

from ..database import my_db


class History(my_db.Base):
    __tablename__ = "history"

    Uid = Column(Integer, ForeignKey("user.Uid"), primary_key=True)
    Pid = Column(Integer, ForeignKey("post.Pid", ondelete='CASCADE'), primary_key=True)
    lastVisitTime = Column(DateTime, default=datetime.datetime.utcnow)

    by_user = relationship("User", back_populates="view")
    related_post = relationship("Post", back_populates="view")

    def __init__(self, Uid, Pid, lastVisitTime=None):
        if isinstance(lastVisitTime, str):
            lastVisitTime = datetime.datetime.strptime(lastVisitTime, "%Y-%m-%d %H:%M:%S")
        self.Uid = Uid
        self.Pid = Pid
        self.lastVisitTime = lastVisitTime

    def __repr__(self):
        return f"<ViewHistory ({self.Uid, self.Pid})>"


import datetime

from sqlalchemy import Column, Integer, String, DateTime, ForeignKey
from sqlalchemy.orm import relationship

from ._tables import user_report_table
from ..database import my_db


class Report(my_db.Base):
    __tablename__ = "report"

    Rid = Column(Integer, primary_key=True)
    target = Column(String)
    targetId = Column(Integer)  # can be Pid or Cid, for simplicity no foreignkey constraints nor relationship
    reason = Column(String)
    timestamp = Column(DateTime, default=datetime.datetime.utcnow)
    resolved = Column(Integer, default=0)  # 0=False, 1=True
    Uid = Column(Integer, ForeignKey("user.Uid"))  # reporter

    report_by = relationship("User", secondary=lambda: user_report_table, back_populates="reports")

    def __init__(self, Uid, target, targetId, reason, timestamp=None, resolved=None):
        if isinstance(timestamp, str):
            timestamp = datetime.datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S")
        self.Uid = Uid
        self.target = target
        self.targetId = targetId
        self.reason = reason
        self.timestamp = timestamp
        self.resolved = resolved

    def __repr__(self):
        return "<Report %r>" % self.Rid


import datetime

from sqlalchemy import Column, Integer, DateTime, ForeignKey
from sqlalchemy.orm import relationship

from ..database import my_db


# insert/update on "like" or "dislike" actions
class CommentStatus(my_db.Base):
    __tablename__ = "comment_status"

    Uid = Column(Integer, ForeignKey("user.Uid"), primary_key=True)
    Cid = Column(Integer, ForeignKey("comment.Cid", ondelete='CASCADE'), primary_key=True)
    liked = Column(Integer, default=0)  # 0 = False, 1 = True
    disliked = Column(Integer, default=0)
    lastModified = Column(DateTime, default=datetime.datetime.utcnow)  # timestamp of last action

    by_user = relationship("User", back_populates="status_comment")
    on_comment = relationship("Comment", back_populates="status_by")

    def __init__(self, Uid, Cid, liked=None, disliked=None, lastModified=None):
        if isinstance(lastModified, str):
            lastModified = datetime.datetime.strptime(lastModified, "%Y-%m-%d %H:%M:%S")
        self.Uid = Uid
        self.Cid = Cid
        self.liked = liked
        self.disliked = disliked
        self.lastModified = lastModified

    def __repr__(self):
        return f"<CommentStatus ({self.Uid}, {self.Cid})>"


import datetime
import hashlib

from sqlalchemy import Column, Integer, String, DateTime

from ..database import my_db


class Admin(my_db.Base):
    __tablename__ = 'admin'

    Aid = Column(Integer, primary_key=True)
    password = Column(String)
    aname = Column(String, unique=True)
    nickname = Column(String)
    avatar = Column(String, default="avatar/default_avatar.jpg")  # upon uploading, link=hash(Aid + timestamp) + ".png"
    timestamp = Column(DateTime, default=datetime.datetime.utcnow)  # time of account creation

    def __init__(self, password, aname=None, nickname=None, avatar=None, timestamp=None):
        if isinstance(timestamp, str):
            timestamp = datetime.datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S")
        self.password = hashlib.sha3_512(password.encode()).hexdigest()
        self.aname = aname if aname else "admin" + str(hash(datetime.datetime.utcnow))
        self.nickname = nickname if nickname else self.aname
        self.avatar = avatar
        self.timestamp = timestamp

    def __repr__(self):
        return '<Admin %r>' % self.Aid


import hashlib
import json
import re
import time

from flask import Blueprint, jsonify, request
from requests_html import HTML

from ..configs import *
from ..database import *
from ..models import *

api = Blueprint("api", __name__, url_prefix="/api")


@api.route('/post/add', methods=["POST"])
@login_required
def add_post():
    """
    This function is used for logged in users to create new post under a board
    :return: json information:
    if the creation is successful it will return status 1 otherwise it will return error message
    """
    Uid = session["Uid"]
    # check whether user is banned
    match_user: User = my_db.query(User, User.Uid == Uid, first=True)
    if match_user.banned:
        if match_user.banDuration > datetime.datetime.utcnow():
            return jsonify({"error": {"msg": "user banned"}, "status": 0})

    Bid = request.form.get("Bid")
    title = request.form.get("title")
    if not Bid or not Bid.isnumeric() or not title:
        return jsonify({"error": {"msg": "invalid data"}, "status": 0})

    if len(title) > 150:
        return jsonify({"error": {"msg": "Title word count exceeded. Maximum: 150"}, "status": 0})

    match_board = my_db.query(Board, Board.Bid == Bid, first=True)
    if not match_board:
        return jsonify({"error": {"msg": "invalid board ID"}, "status": 0})
    match_board.postCount += 1

    content = request.form.get("content", "<p></p>")
    text = request.form.get("text", "")
    if not content:  # unknown bug, the get above does not work
        content = "<p></p>"

    if len(text) > 2000:
        return jsonify({"error": {"msg": "Content word count exceeded. Maximum: 2000"}, "status": 0})

    try:
        html = HTML(html=content)
    except Exception as e:
        return jsonify({"error": {"msg": e}, "status": 0})

    medias = []
    for ele in html.find("img.OT_image,iframe.OT_video"):
        src = ele.attrs.get("src")
        if src:
            tag = ele.tag
            path = PHOTO_PATH if tag == "img" else VIDEO_PATH
            medias.append(path + src.split("/")[-1])

    new_post = Post(Uid, int(Bid), title, content, medias, text)
    my_db.add(new_post)
    return jsonify({"status": 1})


@api.route('/like', methods=["POST"])
@login_required
def like():
    """
    This function is used for logged in users to like a comment or post
    :return: json information:
    if the process is successful it will return status 1 otherwise it will return error message
    """
    Uid = session["Uid"]
    target = request.form.get("target")
    target_id = request.form.get("id")

    if target not in ("comment", "post") or not target_id or not target_id.isnumeric():
        return jsonify({"error": {"msg": "invalid data"}, "status": 0})

    query_from, filter_cond = (Comment, Comment.Cid == target_id) if target == "comment" else (
        Post, Post.Pid == target_id)
    match_target = my_db.query(query_from, filter_cond, first=True)
    if not match_target:
        return jsonify({"error": {"msg": "invalid target ID"}, "status": 0})
    if match_target.status != 0:
        return jsonify({"error": {"msg": "Object not exists"}, "status": 0})

    status = CommentStatus if target == "comment" else PostStatus
    status_cond = CommentStatus.Cid == target_id if target == "comment" else PostStatus.Pid == target_id
    match_status = my_db.query(status, and_(status.Uid == Uid, status_cond), first=True)

    if not match_status:
        cur_status = 1
        new_status = status(Uid, int(target_id), cur_status, 0)
        my_db.add(new_status)
        match_target.likeCount += 1
    else:
        liked = match_status.liked
        disliked = match_status.disliked
        cur_status = 0 if liked else 1
        new_status = status(Uid, int(target_id), cur_status, 0, datetime.datetime.utcnow())
        my_db.merge(new_status)
        match_target.likeCount += -1 if liked else 1
        match_target.dislikeCount -= 1 if disliked else 0

    cur_target = my_db.query(query_from, filter_cond, first=True)
    cur_like, cur_dislike = cur_target.likeCount, cur_target.dislikeCount
    return jsonify({"cur_status": cur_status, "like_count": cur_like, "dislike_count": cur_dislike, "status": 1})


@api.route('/dislike', methods=["POST"])
@login_required
def dislike():
    """
    This function is used for logged in users to dislike a comment or post
    :return: json information:
    if the process is successful it will return status 1 otherwise it will return error message
    """
    Uid = session["Uid"]
    target = request.form.get("target")
    target_id = request.form.get("id")

    if target not in ("comment", "post") or not target_id or not target_id.isnumeric():
        return jsonify({"error": {"msg": "invalid data"}, "status": 0})

    query_from, filter_cond = (Comment, Comment.Cid == target_id) if target == "comment" else (
        Post, Post.Pid == target_id)
    match_target = my_db.query(query_from, filter_cond, first=True)
    if not match_target:
        return jsonify({"error": {"msg": "invalid target ID"}, "status": 0})
    if match_target.status != 0:
        return jsonify({"error": {"msg": "Object not exists"}, "status": 0})

    status = CommentStatus if target == "comment" else PostStatus
    status_cond = CommentStatus.Cid == target_id if target == "comment" else PostStatus.Pid == target_id
    match_status = my_db.query(status, and_(status.Uid == Uid, status_cond), first=True)

    if not match_status:
        cur_status = 1
        new_status = status(Uid, int(target_id), 0, 1)
        my_db.add(new_status)
        match_target.dislikeCount += 1
    else:
        liked = match_status.liked
        disliked = match_status.disliked
        cur_status = 0 if disliked else 1
        new_status = status(Uid, int(target_id), 0, cur_status, datetime.datetime.utcnow())
        my_db.merge(new_status)
        match_target.dislikeCount += -1 if disliked else 1
        match_target.likeCount -= 1 if liked else 0

    cur_target = my_db.query(query_from, filter_cond, first=True)
    cur_like, cur_dislike = cur_target.likeCount, cur_target.dislikeCount
    return jsonify({"cur_status": cur_status, "like_count": cur_like, "dislike_count": cur_dislike, "status": 1})


@api.route('/report/add', methods=["POST"])
@login_required
def add_report():
    """
    This function is used for logged in users to report a post
    :return: if the report is successful, it will redirect the user to the previous post page
    otherwise, it will return json error message
    """
    Uid = session["Uid"]
    target = request.form.get("target")
    target_id = request.form.get("id")
    reason = request.form.get("reason")
    if target not in ["comment", "post"] or not target_id or not target_id.isnumeric() or not reason:
        return jsonify({"error": {"msg": "invalid data"}}), 403

    query_from, filter_cond = (Comment, Comment.Cid == target_id) if target == "comment" else (
        Post, Post.Pid == target_id)
    match_target = my_db.query(query_from, filter_cond, first=True)
    if not match_target:
        return jsonify({"error": {"msg": "invalid target ID"}}), 403

    Pid = match_target.Pid
    # insert into db
    new_report = Report(Uid, target, int(target_id), reason)
    my_db.add(new_report)

    reporter = my_db.query(User, User.Uid == Uid, first=True)
    reporter.reports.append(new_report)
    return redirect(f"/post/{Pid}")


@api.route('/comment/add', methods=["POST"])
@login_required
def add_comment():
    """
    This function is used for logged in users to create new comment under a post
    :return: json information:
    if the creation is successful it will return status 1 otherwise it will return error message
    """
    Uid = session["Uid"]
    # check whether user is banned
    match_user: User = my_db.query(User, User.Uid == Uid, first=True)
    if match_user.banned:
        if match_user.banDuration > datetime.datetime.utcnow():
            return jsonify({"error": {"msg": "User banned."}, "status": 0})

    # verify post data in correct format
    Pid = request.form.get("Pid")
    content = request.form.get("content")
    if not Pid or not Pid.isnumeric() or not content:
        return jsonify({"error": {"msg": "Invalid data."}, "status": 0})

    # verify post exists
    match_post = my_db.query(Post, and_(Post.Pid == Pid, Post.status == 0), first=True)
    if not match_post:
        return jsonify({"error": {"msg": "Post not found."}, "status": 0})

    # verify text not too long
    text = request.form.get("text", "")  # can be None because comment may only contain photos and/or videos
    if len(text) > 1000:
        return jsonify({"error": {"msg": "Word count exceeded. Maximum: 1000"}, "status": 0})

    # make content HTML for parsing
    try:
        html = HTML(html=content)
    except Exception as e:
        return jsonify({"error": {"msg": e}, "status": 0})

    # fetch media list
    medias = []
    for ele in html.find("img.OT_image,iframe.OT_video"):
        src = ele.attrs.get("src")
        if src:
            tag = ele.tag
            path = PHOTO_PATH if tag == "img" else VIDEO_PATH
            medias.append(path + src.split("/")[-1])

    # check if the comment is replying other's comment
    reply_ele = html.find(".OT_reply", first=True)
    if reply_ele:
        ele_text_length = len(reply_ele.text)
        if len(text) <= ele_text_length and not medias:  # content is empty if not text nor media
            return jsonify({"error": {"msg": "Empty reply!"}, "status": 0})

        # retrieve receiver ID and target ID from data attributes (added by ourself) in HTML tag
        Rid = reply_ele.attrs["data-uid"]
        Tid = reply_ele.attrs["data-cid"]
        # if sender != receiver, send notification to comment owner
        if Uid != Rid:
            ntf_to_commenter = Notification("user", Uid, "user", Rid, "comment", Tid, "reply")
            my_db.add(ntf_to_commenter)

    # if sender != receiver, send notification to post owner
    if Uid != (Rid := match_post.owner.Uid):
        ntf_to_poster = Notification("user", Uid, "user", Rid, "post", Pid, "comment")
        my_db.add(ntf_to_poster)

    # record current available floor
    floor = match_post.available_floor
    # update post statistics
    match_post.available_floor += 1
    match_post.commentCount += 1
    match_post.latestCommentTime = datetime.datetime.utcnow()

    # add comment into database
    new_comment = Comment(Uid, Pid, content, floor, medias, text)
    my_db.add(new_comment)
    return jsonify({"status": 1})


@api.route('/post/delete', methods=["POST"])
@login_required
def delete_post():
    """
    THis function is used for logged in users to delete a post created by himself
    :return: if successful, it will redirect to the previous board
    otherwise, it will return json error message
    """
    Uid = session['Uid']

    Pid = request.form.get("Pid")
    if not Pid or not Pid.isnumeric():
        return jsonify({"error": {"msg": "Invalid data."}, "status": 0})

    match_post = my_db.query(Post, and_(Post.Pid == Pid, Post.Uid == Uid, Post.status == 0), first=True)
    if not match_post:
        return jsonify({"error": {"msg": "Post not found."}, "status": 0})

    match_post.under.postCount -= 1
    my_db.update(Post, Post.Pid == Pid, values={"status": 1})

    # Then delete all corresponding data in other relating tables
    for c in match_post.comments:
        my_db.update(Comment, Comment.Pid == c.Pid, values={"status": 1})

    return jsonify({"status": 1})


@api.route('/comment/delete', methods=["POST"])
@login_required
def delete_comment():  # will not alter post lastCommentTime
    """
    THis function is used for logged in users to delete a comment created by himself
    :return: if successful, it will redirect to the previous post
    otherwise, it will return json error message
    """
    Uid = session['Uid']

    Cid = request.form.get("Cid")
    if not Cid or not Cid.isnumeric():
        return jsonify({"error": {"msg": "Invalid data."}, "status": 0})
    match_comment = my_db.query(Comment, and_(Comment.Cid == Cid, Comment.Uid == Uid, Comment.status == 0), first=True)
    if not match_comment:
        return jsonify({"error": {"msg": "Comment not found."}, "status": 0})

    match_comment.comment_in.commentCount -= 1
    my_db.update(Comment, Comment.Cid == Cid, values={"status": 1})

    return jsonify({"status": 1})


@api.route('/post/restore', methods=["POST"])
@login_required
def restore_post():
    Uid = session['Uid']

    Pid = request.form.get("Pid")
    if not Pid or not Pid.isnumeric():
        return jsonify({"error": {"msg": "Invalid data."}, "status": 0})

    match_post = my_db.query(Post, and_(Post.Pid == Pid, Post.Uid == Uid, Post.status == 1), first=True)
    if not match_post:
        return jsonify({"error": {"msg": "Post not found."}, "status": 0})

    match_post.under.postCount += 1
    my_db.update(Post, Post.Pid == Pid, values={"status": 0})

    # Then restore all corresponding data in other relating tables
    for c in match_post.comments:
        my_db.update(Comment, Comment.Pid == c.Pid, values={"status": 0})

    return jsonify({"status": 1})


@api.route('/comment/restore', methods=["POST"])
@login_required
def restore_comment():
    Uid = session['Uid']

    Cid = request.form.get("Cid")
    if not Cid or not Cid.isnumeric():
        return jsonify({"error": {"msg": "Invalid data."}, "status": 0})
    match_comment = my_db.query(Comment, and_(Comment.Cid == Cid, Comment.Uid == Uid, Comment.status == 1), first=True)
    if not match_comment:
        return jsonify({"error": {"msg": "Comment not found."}, "status": 0})

    match_comment.comment_in.commentCount += 1
    my_db.update(Comment, Comment.Cid == Cid, values={"status": 0})

    return jsonify({"status": 1})


@api.route('/personal_info/add', methods=["POST"])
@login_required
def add_personal_info():
    """
    This function is used for logged in users to add information in their profile
    :return: if successful, it will redirect to the profile page of this user
    otherwise, it will return json error message
    """
    Uid = session["Uid"]

    nickname = request.form.get("nickname")
    if not nickname:
        return jsonify({"error": {"msg": "invalid data"}}), 403
    # check gender
    gender = request.form.get("gender")
    if gender not in ["male", "female", "other"]:
        return jsonify({"error": {"msg": "invalid gender"}}), 403
    # check phone number
    phone_number = request.form.get("phone_number")
    phone_number = re.findall(r"^[+]*[(]?[0-9]{1,4}[)]?[-\s./0-9]*$", phone_number)
    if not phone_number:
        return jsonify({"error": {"msg": "invalid phone number"}}), 403
    else:
        phone_number = phone_number[0]
    # check email
    email = request.form.get("email")
    email = re.findall(r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$", email)
    if not email:
        return jsonify({"error": {"msg": "invalid email"}}), 403
    else:
        email = email[0]
    # check address
    address = request.form.get("address")
    if len(address) > 200:
        return jsonify({"error": {"msg": "invalid address"}}), 403
    # check date of birth
    date_of_birth = request.form.get("date_of_birth")
    try:
        date_of_birth = datetime.datetime.strptime(date_of_birth, "%Y-%m-%d")
    except Exception as e:
        return jsonify({"error": {"msg": f"invalid date of birth: {e}"}}), 403

    my_db.update(User, User.Uid == Uid, values={"nickname": nickname, "gender": gender, "phoneNumber": phone_number,
                                                "email": email, "address": address, "dateOfBirth": date_of_birth})
    match_user = my_db.query(User, User.Uid == Uid, first=True)
    session.pop("user_info")
    session["user_info"] = {"nickname": match_user.nickname, "avatar": match_user.avatar}
    return redirect(f"/profile/{Uid}")


@api.route('/auth/register', methods=["POST"])
def register_auth():
    """
    This function is to evaluate the registration
    :return: json information:
    if the registration is successful it will return status 1 otherwise it will return error message
    """
    Uid = session.get("Uid")
    if Uid:
        return redirect("/")  # if already logged in, redirect to homepage

    # check username
    username = request.form.get("uname")
    if not username:
        return jsonify({"error": {"msg": "Invalid data"}, "status": 0})
    # (non-existence)
    match_user = my_db.query(User, User.uname == username, first=True)
    if match_user:
        return jsonify({"error": {"msg": "user already exists"}, "status": 0})
    # (validity)
    username = re.findall(r"[\w_]+$", username)
    if not username:
        return jsonify({"error": {"msg": "Invalid username"}, "status": 0})
    else:
        username = username[0]
    if len(username) < 5 or len(username) > 20:
        return jsonify({"error": {"msg": "username must be of length 5 ~ 20"}, "status": 0})
    # check password
    password = request.form.get("password")
    if not password:
        return jsonify({"error": {"msg": "Invalid data"}, "status": 0})
    password = re.findall(r"^(?=.*[A-Za-z])(?=.*\d)(?=.*[@$!%*#?&])[A-Za-z\d@$!%*#?&]{8,}$", password)
    if not password:
        return jsonify({"error": {"msg": "Invalid password"}, "status": 0})
    else:
        password = password[0]
    # check nickname
    nickname = request.form.get("nickname")
    if not nickname or len(nickname) > 20:
        return jsonify({"error": {"msg": "Invalid nickname"}, "status": 0})

    new_user = User(password, username, nickname=nickname)
    my_db.add(new_user)

    # login once finish registration
    new_user = my_db.query(User, User.uname == username, first=True)
    session["Uid"] = new_user.Uid
    session["user_info"] = {"nickname": new_user.nickname, "avatar": new_user.avatar}
    return jsonify({"status": 1})


@api.route('/auth/login', methods=["POST"])
def login_auth():
    """
    This function is to evaluate the log in
    :return: json information:
    if the log in is successful it will return status 1 and corresponding Uid otherwise it will return error message
    """
    Uid = session.get("Uid")
    if Uid:
        return redirect("/")  # if already logged in, redirect to homepage

    data = request.form.to_dict()
    username = data.get("uname")
    password = data.get("password")
    if not username or not password:
        return jsonify({"error": {"msg": "Invalid input."}, "status": 0})

    match_user: User = my_db.query(User, User.uname == username, first=True)
    if not match_user:
        return jsonify({"error": {"msg": "Username does not exist."}, "status": 0})
    if hashlib.sha3_512(password.encode()).hexdigest() != match_user.password:
        return jsonify({"error": {"msg": "Incorrect password."}, "status": 0})
    session["Uid"] = match_user.Uid
    user_info = {"nickname": match_user.nickname, "avatar": match_user.avatar}
    session["user_info"] = user_info
    session["last_check"] = match_user.lastCheck
    # session.permanent = True
    return jsonify({"status": 1, "Uid": match_user.Uid})


@api.route('/auth/logout', methods=["POST", "GET"])
@login_required
def logout_auth():
    """
    This function is used for logged in user to logout
    :return: a html text message
    """
    session.clear()
    return "<script>location.replace(document.referrer);</script>", 200


@api.route("/upload", methods=["POST", "GET"])
def handle_upload():
    """
    This function is used for users to upload their avatar, image or video
    :return: if this user is not logged in, it will redirect to the log in page
    if the uploading failed, it will return json error message
    """
    action = request.args.get("action")
    method = request.method.upper()

    if action == "uploadavatar" and method == "POST":  # user action. Upload avatar
        if not session.get("Uid"):
            return redirect("/login")

        Uid = session["Uid"]
        file = request.files.get("file")
        # check if file
        if not file:
            return jsonify({"error": {"msg": "Please upload a file"}})
        # check file type
        file_type = file.content_type
        if not file_type or not file_type.startswith("image"):
            return jsonify({"error": {"msg": "Invalid file type"}})
        file_type = file_type.split("/")[-1]
        # check file size
        file_size = int(request.headers.get("Content-Length", 0))
        if file_size > 3 * 1024 * 1024:
            return jsonify({"error": {"msg": "Image too large"}})

        path = CDN_ROOT_PATH + AVATAR_PATH
        if not os.path.exists(path):  # os is imported in config.py
            os.mkdir(path)

        src = str(hash(str(Uid) + str(datetime.datetime.utcnow()))) + "." + file_type
        filepath = path + src
        while os.path.exists(filepath):
            src = str(hash(str(Uid) + str(datetime.datetime.utcnow()))) + "." + file_type
            filepath = path + src
        file.save(filepath)

        match_user = my_db.query(User, User.Uid == Uid, first=True)
        avatar = match_user.avatar
        if avatar != "default_avatar.jpg":
            old_path = path + avatar
            if os.path.exists(old_path):
                os.remove(old_path)

        new_avatar = AVATAR_PATH + src
        my_db.update(User, User.Uid == Uid, values={"avatar": new_avatar})
        session.pop("user_info")
        session["user_info"] = {"nickname": match_user.nickname, "avatar": new_avatar}
        result = {"status": 1}

    elif action == "config" and method == "GET":  # ueditor action. Config the ueditor, user may not be logged-in
        with open("ourtieba/static/ueditor/config.json", "r") as f:
            content = f.read()
        result = json.loads(content)

    elif action == "uploadimage" and method == "POST":  # ueditor + user action. Upload photo within post and comment
        if not session.get("Uid"):
            return jsonify({"error": {"msg": "Not logged in"}, "status": 0})  # AE: can be any error, ignored by ueditor
        Uid = session["Uid"]

        file = request.files.get("upfile")
        file_type = file.content_type
        if not file_type or not file_type.startswith("image"):
            return jsonify({"error": {"msg": "Invalid file type"}})
        file_type = file_type.split("/")[-1]

        file_size = int(request.headers.get("Content-Length", 0))
        if file_size > 2048000:  # must be the same as in static/ueditor/config.json ("imageMaxSize")
            return jsonify({"error": {"msg": "Not logged in"}, "status": 0})  # AE

        path = CDN_ROOT_PATH + PHOTO_PATH
        if not os.path.exists(path):  # os is imported in config.py
            os.mkdir(path)

        src = str(hash(str(Uid) + str(datetime.datetime.utcnow()))) + "." + file_type
        filepath = path + src
        while os.path.exists(filepath):
            src = str(hash(str(Uid) + str(datetime.datetime.utcnow()))) + "." + file_type
            filepath = path + src
        file.save(filepath)

        result = {
            "state": "SUCCESS",
            "url": "/" + filepath,
            "title": "",
            "original": ""
        }

    elif action == "uploadvideo" and method == "POST":  # ueditor + user action. Upload video within post and comment
        if not session.get("Uid"):
            return jsonify({"error": {"msg": "Not logged in"}, "status": 0})
        Uid = session["Uid"]

        file = request.files.get("upfile")
        file_type = file.content_type
        if not file_type or not file_type.startswith("video"):
            return jsonify({"error": {"msg": "Invalid file type"}})
        file_type = file_type.split("/")[-1]

        file_size = int(request.headers.get("Content-Length", 0))
        if file_size > 102400000:  # must be the same as in static/ueditor/config.json ("videoMaxSize")
            return jsonify({"error": {"msg": "Not logged in"}, "status": 0})  # AE

        path = CDN_ROOT_PATH + VIDEO_PATH
        if not os.path.exists(path):  # os is imported in config.py
            os.mkdir(path)

        src = str(hash(str(Uid) + str(datetime.datetime.utcnow()))) + "." + file_type
        filepath = path + src
        while os.path.exists(filepath):
            src = str(hash(str(Uid) + str(datetime.datetime.utcnow()))) + "." + file_type
            filepath = path + src
        file.save(filepath)

        result = {
            "state": "SUCCESS",
            "url": "/" + filepath,
            "title": "",
            "original": ""
        }
    else:
        result = {"error": {"msg": "Something went wrong"}, "status": 0}
    return jsonify(result)


@api.route("/subscribe", methods=["POST"])
@login_required
def subscribe():
    """
    This function is used for logged in users to subscribe a board
    :return: json information
    if the subscribe is successful it will return status 1 otherwise it will return error message
    """
    Uid = session["Uid"]

    Bid = request.form.get("Bid")
    action = request.form.get("action")  # "0"=unsub, "1"=sub
    if not Bid or not Bid.isnumeric() or action not in ("0", "1"):
        return jsonify({"error": {"msg": "invalid data"}, "status": 0})

    match_board = my_db.query(Board, Board.Bid == Bid, first=True)
    if not match_board:
        return jsonify({"error": {"msg": "invalid board ID"}, "status": 0})
    if match_board.status != 0:
        return jsonify({"error": {"msg": "Board not exists"}, "status": 0})
    match_board.subscribeCount += 1 if action == "1" else -1

    new_sub = Subscription(Uid, Bid, int(action), lastModified=datetime.datetime.utcnow())
    my_db.merge(new_sub)
    return jsonify({"subs_count": match_board.subscribeCount, "status": 1})


@api.route("/auth/set_password")
@login_required
def set_password():
    """
    This function is used for logged in users to evaluate password setting process
    :return: json information
    if the setting is successful it will return status 1 otherwise it will return error message
    """
    Uid = session["Uid"]

    # check password
    password = request.form.get("password")
    if not password:
        return jsonify({"error": {"msg": "Invalid data"}, "status": 0})
    password = re.findall(r"^(?=.*[A-Za-z])(?=.*\d)(?=.*[@$!%*#?&])[A-Za-z\d@$!%*#?&]{8,}$", password)
    if not password:
        return jsonify({"error": {"msg": "Invalid password"}, "status": 0})
    else:
        password = hashlib.sha3_512(password[0].encode()).hexdigest()

    my_db.update(User, User.Uid == Uid, values={"password": password})
    return jsonify({"status": 1})


@api.route("/fetch")
def fetch_data():
    cur_Uid = session.get("Uid", 0)
    # Optional: Can block user's request when cur_Uid != session["Uid"] (the user is viewing other's profile)

    Uid = request.args.get("Uid")
    type_data = request.args.get("type")

    if not Uid or not type_data or not type_data.isnumeric() or not (0 <= (type_data := int(type_data)) <= 3):
        return jsonify({"error": {"msg": "Invalid data!"}, "status": 0})

    match_user = my_db.query(User, User.Uid == Uid, first=True)
    if not match_user:
        return jsonify({"error": {"msg": "No user found!"}, "status": 0})

    base_info = {"status": 1}

    if type_data == 0:
        post_info = [{"Pid": p.Pid, "Bid": p.Bid, "bname": p.under.name, "title": p.title,
                      "timestamp": p.timestamp, "status": p.status} for p in match_user.posts]
        # sort by timestamp desc
        post_info.sort(key=lambda p: p["timestamp"], reverse=True)
        # convert times into shorter format
        for p in post_info:
            p["timestamp"] = convert_time(p["timestamp"])
        base_info.update({"info": post_info, "count": len(post_info)})
    elif type_data == 3:
        history_info = []
        for h in match_user.view:
            history = {"Pid": h.Pid, "LVT": h.lastVisitTime}
            p = h.related_post
            history.update({"title": p.title, "bname": p.under.name, "Bid": p.Bid, "Uid": (u := p.owner).Uid,
                            "nickname": u.nickname, "me": int(u.Uid == cur_Uid),
                            "status": p.status})  # "me" = whether post by me
            history_info.append(history)
        # sort by LVT desc
        history_info.sort(key=lambda ht: ht["LVT"], reverse=True)
        # convert times into shorter format
        for h in history_info:
            h["LVT"] = convert_time(h["LVT"])
        base_info.update({"info": history_info, "count": len(history_info)})
    elif type_data == 2:
        subs_info = []
        for s in match_user.subscriptions:
            if s.subscribed == 1:
                if (b := s.of_board).status == 0:
                    subs_info.append({"Bid": s.Bid, "bname": b.name, "LM": s.lastModified,
                                      "cover": b.cover, "status": 0})
        # sort by LM desc
        subs_info.sort(key=lambda sb: sb["LM"], reverse=True)
        base_info.update({"info": subs_info, "count": len(subs_info)})
    else:
        comment_info = []
        for c in match_user.comments:
            comment = {"Cid": c.Cid, "text": c.text, "timestamp": c.timestamp, "status": c.status}
            p = c.comment_in
            comment.update({"Pid": p.Pid, "title": p.title})
            comment_info.append(comment)
        # sort by timestamp desc
        comment_info.sort(key=lambda ci: ci["timestamp"], reverse=True)
        # convert times into shorter format
        for c in comment_info:
            c["timestamp"] = convert_time(c["timestamp"])
        base_info.update({"info": comment_info, "count": len(comment_info)})

    return jsonify(base_info)


@api.route("/get_log")
def get_log():
    Uid = session.get("Uid")
    last_check = session.get("last_check")
    if not Uid or not last_check:
        return jsonify({"code": -1})
    cur_ts = request.args.get("t") or time.time()  # can be used for synchronization given FIFO channel
    new_count = my_db.count(Notification, and_(Notification.receiver == "user", Notification.Rid == Uid,
                                               Notification.timestamp.between(last_check, cur_ts)))
    if not new_count:
        return jsonify({"code": 204})  # empty response
    return jsonify({"code": 200, "new_count": new_count})



import time

from sqlalchemy import Column, Integer, String, DECIMAL

from ..database import my_db


class Notification(my_db.Base):
    __tablename__ = "notification"

    Nid = Column(Integer, primary_key=True)
    # All the columns are NOT nullable!
    starter = Column(String)  # "user", "admin"
    Sid = Column(Integer)  # starter's Uid or Aid
    receiver = Column(String)  # "user", "broadcast"(not implemented)
    Rid = Column(Integer)  # receiver's Uid, if broadcast set to -1
    target = Column(String)  # "post", "comment", "user", "board"(not implemented)
    Tid = Column(Integer)  # target's Pid or Cid or Uid or Bid
    action = Column(String)  # "like", "dislike", "comment", "delete", "restore", "ban", "unban", "post"(not implement)
    timestamp = Column(Integer, default=time.time)  # must be UNIX timestamp, like "1620750628.290329" etc.

    def __init__(self, starter, Sid, receiver, Rid, target, Tid, action, timestamp=None):
        self.starter = starter
        self.Sid = Sid
        self.receiver = receiver
        self.Rid = Rid
        self.target = target
        self.Tid = Tid
        self.action = action
        self.timestamp = timestamp


from ..views.abstract_user import *
from ..views.admin import *
from ..views.api import *
from ..views.cdn import *
from ..views.guest import *
from ..views.user import *


def register_blue(app):
    app.register_blueprint(user_blue)
    app.register_blueprint(guest)
    app.register_blueprint(api)
    app.register_blueprint(admin_blue)
    app.register_blueprint(a_user)
    app.register_blueprint(cdn)



from setuptools import find_packages, setup

setup(
    name='ourtieba',
    version='0.1.0',
    packages=find_packages(),
    license='MIT',
    description='simple blog implementation',
    url='https://github.com/nyush-se-spring2021-forum/OurTieba',
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'flask',
        'sqlalchemy',
        'flask-moment',
        'flask_apscheduler',
        'requests_html'
    ]
)



from flask_apscheduler import APScheduler

from .logger import *
from .models import *

# logger = init_logger()

record = {}

def update_hot():
    """ Not Finished!"""
    global record
    print("updating hot...")
    # logger.logger.info("Start updating hot")
    all_boards = my_db.query(Board, True)  #get all boards
    #Firstly we need to check all Bid and delete non-existing Bid in record
    Bid_list = []
    for j in all_boards:
        Bid_list.append(j.Bid)
    for k in record.keys():
        if k not in Bid_list:
            record.pop(k)
    #Then we start to do the updating:
    for i in all_boards:
        Bid = i.Bid
        old_list = record.get(Bid, None)
        if old_list is None:
            old_list = [0, 0]
            record[Bid] = old_list
        #Algorithm of calculating a board's hot:
        #50% * new_postCount + 50% * new_viewCount
        new_postCount = i.postCount - old_list[0]
        new_viewCount = i.viewCount - old_list[1]
        new_hot = 0.5 * new_postCount + 0.5 * new_viewCount

        #Then update the record dic
        new_list = [i.postCount, i.viewCount]
        record[Bid] = new_list
        my_db.update(Board, Board.Bid == Bid, values={"hot": new_hot})

    print("average hot:", my_db.query(func.avg(Board.hot).label("average")).scalar())
    my_db.commit()


def init_scheduler(app):
    scheduler = APScheduler()
    scheduler.add_job(func=update_hot, id="1", trigger="interval", seconds=86400)  # trigger everyday
    scheduler.init_app(app)
    scheduler.start()



# Macros
PAGE_SIZE = 10
RECOMMEND_NUM_BOARD = 10
RECOMMEND_NUM_NEWS = 3
NEWS_UPDATE_FREQUENCY = 60 * 60  # unit: second

CDN_ROOT_PATH = "cdn/"  # relative to ourtieba working directory (which is .../OurTieba)
AVATAR_PATH = "avatar/"  # relative to CDN root path
PHOTO_PATH = "photo/"  # relative to CDN root path
VIDEO_PATH = "video/"  # relative to CDN root path

ALLOWED_METHODS = ("GET", "POST")



import time

from flask import Blueprint, jsonify, render_template, request

from ..configs.functions import *
from ..database import *
from ..models import *

user_blue = Blueprint("user", __name__)


@user_blue.route("/board/create")
@login_required
def create_post():
    """
    not used
    :return:
    """
    Uid = session["Uid"]

    # check whether user is banned
    match_user: User = my_db.query(User, User.Uid == Uid, first=True)
    if match_user.banned:
        if match_user.banDuration > datetime.datetime.utcnow():
            return jsonify({"error": {"msg": "user banned"}}), 404
    Bid = request.args.get("Bid")
    if not Bid or not Bid.isnumeric():
        return jsonify({"error": {"msg": "invalid data"}}), 404

    match_board = my_db.query(Board, Board.Bid == Bid)
    if not match_board:
        return jsonify({"error": {"msg": "invalid board ID"}}), 404
    data = {"Bid": Bid}
    return render_template("login.html", data=data)


@user_blue.route("/report")
@login_required
def report():
    """
    This function is used to redirect the users to the report page
    :return: report.html
    """
    target = request.args.get("target", 0)
    id = request.args.get("id")
    data = {"id": id, "target": target}
    if target == "comment":
        match_result = my_db.query(Comment, Comment.Cid == id, first=True)
        if not match_result:
            return "Not Found", 404
        return render_template("report.html", data=data)
    elif target == "post":
        match_result = my_db.query(Post, Post.Pid == id, first=True)
        if not match_result:
            return "Not Found", 404
        return render_template("report.html", data=data)
    else:
        return "Invalid URL", 404


@user_blue.route("/notifications")
@login_required
def check_notification():
    Uid = session["Uid"]
    last_check = session["last_check"]

    cur_ts = time.time()
    match_ntf = my_db.query(Notification, and_(Notification.receiver == "user", Notification.Rid == Uid,
                                               Notification.timestamp.between(last_check, cur_ts)))
    ntfs = []
    for n in match_ntf:
        n: Notification
        ntfs.append({"starter": n.starter, "Sid": n.Sid, "target": n.target, "Tid": n.Tid,
                     "action": n.action, "timestamp": n.timestamp})
    data = {"ntfs": ntfs}
    my_db.update(User, User.Uid == Uid, values={"lastCheck": cur_ts})
    session["last_check"] = cur_ts
    return render_template("notifications.html", data=data)


import datetime

from requests_html import HTMLSession, AsyncHTMLSession


class OTSpider:
    __instance = None

    def __new__(cls, *args, **kwargs):
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance

    def __init__(self):
        self.session = HTMLSession()
        self.headers = {"User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, "
                                      "like Gecko) Chrome/89.0.4389.90 Safari/537.36 Edg/89.0.774.57"}
        self.cache = {"news": dict()}  # note that this is stateless, which means it is cleared on every restart

    def get(self, *arg, **kwargs):
        return self.session.get(*arg, **kwargs)

    def post(self, *arg, **kwargs):
        return self.session.post(*arg, **kwargs)

    def set_cookie(self, cookie):  # if cookie set to None, remove cookie
        self.headers.update({"Cookie": cookie})
        if not cookie:
            self.headers.pop("Cookie")

    def empty_cache(self, name=None):  # empty all the cache by default
        if not name:
            self.cache = dict()
            return
        if not self.cache.get(name):
            print("Name not found! Aborting...")
            return
        self.cache.pop(name)

    def get_hot_news(self, **kwargs):
        country = kwargs.get('country', 'us') or "us"  # (no "country" in kwargs) or (country=None)
        category = kwargs.get('category', 'general') or "general"
        # if news is not cached, or cache has expired, update cache
        if not (news := self.cache["news"].get(key := (country, category))) or \
                news["expires"] < datetime.datetime.utcnow():
            url = f"https://newsapi.org/v2/top-headlines?country={country}&category={category}&" \
                  "apiKey=9d51c192354848748d129b08faea32ea"
            res = self.session.get(url, headers=self.headers)
            data = res.json()
            num = kwargs["num"]
            self.cache["news"][key] = {"articles": data["articles"][:num],
                                       "expires": datetime.datetime.utcnow() + datetime.timedelta(
                                           seconds=kwargs["freq"])}
        return self.cache["news"][key]["articles"]

    def weibo_hot_search(self, **kwargs):  # cache not applied because change is rapid
        # Juncheng's cookie
        self.set_cookie('SINAGLOBAL=9266532810528.248.1599373736944; login_sid_t=b3eea3454137ac70848c4a32a25c282a; '
                        'cross_origin_proto=SSL; _s_tentry=-; Apache=349928957021.5636.1618803988393; '
                        'ULV=1618803988399:9:2:1:349928957021.5636.1618803988393:1617765463404; '
                        'wb_view_log=1536*8641.25&1920*10802.0000000298023224; ALF=1650340136; '
                        'SSOLoginState=1618804137; '
                        'SUB'
                        '=_2A25NeI38DeRhGeFN71AZ9y7Nwj2IHXVuD_g0rDV8PUNbmtANLVLmkW9NQDBJhXO328RvICGNnS19JPY6LqASfZKJ; '
                        'SUBP=0033WrSXqPxfM725Ws9jqgMF55529P9D9WhfZk1zy.rnRggXMDx9zNI55JpX5KzhUgL'
                        '.FoM0ShzRS05p1K22dJLoIp7LxKML1KBLBKnLxKqL1hnLBoMNe0BE1hM7eK.p; wvr=6; '
                        'wb_view_log_7342870191=1536*8641.25&1920*10802.0000000298023224; UOR=www.baidu.com,'
                        'weibo.com,www.baidu.com; webim_unReadCount={"time":1618804704383,"dm_pub_total":0,'
                        '"chat_group_client":0,"chat_group_notice":0,"allcountNum":11,"msgbox":0}')
        res = self.session.get("https://s.weibo.com/top/summary?Refer=top_hot&topnav=1&wvr=6", headers=self.headers)
        selectors = [f"#pl_top_realtimehot > table > tbody > tr:nth-child({i}) > td.td-02 > a"
                     for i in range(2, 2 + int(kwargs["num"]))]
        top_list = []
        for s in selectors:
            element = res.html.find(s, first=True)
            top_list.append({"url": "https://s.weibo.com" + element.attrs["href"], "title": element.text})
        return top_list

    def get_cookie(self, domain="localhost", port=80, uname="U1", password="111"):
        cached_cookies = self.cache.get("cookies")
        if cached_cookies:
            if cached_cookies.get(uname):
                return cached_cookies[uname]

        url = f"http://{domain}:{port}/api/auth/login"
        res = self.session.post(url, data={"uname": uname, "password": password}, headers=self.headers)
        cookie = res.headers["Set-Cookie"].split(";")[0]

        if not cached_cookies:
            self.cache["cookies"] = dict()
        self.cache["cookies"].update({uname: cookie})
        return cookie

    def upload_avatar(self, domain="localhost", port=80, uname="U1"):
        url = f"http://{domain}:{port}/api/upload"
        cookie = self.get_cookie(domain=domain, port=port, uname=uname)
        self.set_cookie(cookie)
        file = ("test", open("test.jpg", "rb"), "image/jpeg")  # name, filename, content_type
        res = self.session.post(url, files={"file": file}, headers=self.headers)
        status = res.json().get("status")
        if status:
            print("Upload Successful!")

    def close(self):
        self.session.close()


class AsyncOTSpider:
    __instance = None

    def __new__(cls, *args, **kwargs):
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance

    def __init__(self):
        self.session = AsyncHTMLSession()
        self.headers = {"User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, "
                                      "like Gecko) Chrome/89.0.4389.90 Safari/537.36 Edg/89.0.774.57"}

    def set_cookie(self, cookie):
        self.headers.update({"Cookie": cookie})

    def remove_cookie(self):
        if self.headers.get("Cookie"):
            self.headers.pop("Cookie")

    async def get_baidu(self):
        res = await self.session.get("https://www.baidu.com", headers=self.headers)
        return res.html.text

    async def get_bilibili(self):
        res = await self.session.get("https://www.bilibili.com", headers=self.headers)
        return res.html.text

    def run_multitask(self, *args, **kwargs):  # can take in "tasks"(Iterable) as keyword argument
        if kwargs.get("tasks"):
            results = self.session.run(*kwargs["tasks"])
        else:
            results = self.session.run(*args)
        return results


class scrapperFactory:

    @staticmethod
    def produce(sync=True):
        return OTSpider() if sync else AsyncOTSpider()


OT_spider = scrapperFactory.produce()

if __name__ == '__main__':
    # news = OT_spider.get_hot_news(num=3, freq=60)
    # print(news)
    # print(async_spider.run_multitask(async_spider.get_baidu))
    # OT_spider.upload_avatar(port=5000)
    print(OT_spider.weibo_hot_search(num=10))
